
from .tableentrysort import sort_table_entries
from .common import assert_prettifier_works


def test_table_sorting():
    toml_text = """description = ""
firstname = "adnan"
lastname = "fatayerji"
git_aydo = ""
groups = ["sales", "dubai", "mgmt"]
skype = ""
emails = ["adnan@incubaid.com",
 "fatayera@incubaid.com",
 "adnan.fatayerji@incubaid.com",
 "adnan@greenitglobe.com",
 "fatayera@greenitglobe.com",
 "adnan.fatayerji@greenitglobe.com"]
# I really like this table
id = "fatayera"
git_github = ""
telegram = "971507192009"
mobiles = ["971507192009"]
"""

    prettified = """description = ""
emails = ["adnan@incubaid.com",
 "fatayera@incubaid.com",
 "adnan.fatayerji@incubaid.com",
 "adnan@greenitglobe.com",
 "fatayera@greenitglobe.com",
 "adnan.fatayerji@greenitglobe.com"]
firstname = "adnan"
git_aydo = ""
git_github = ""
groups = ["sales", "dubai", "mgmt"]
# I really like this table
id = "fatayera"
lastname = "fatayerji"
mobiles = ["971507192009"]
skype = ""
telegram = "971507192009"
"""

    assert_prettifier_works(toml_text, prettified, sort_table_entries)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import codecs
import os
import sys
from shutil import rmtree

from setuptools import find_packages, setup, Command

here = os.path.abspath(os.path.dirname(__file__))
with codecs.open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = '\n' + f.read()
about = {}
with open(os.path.join(here, "pipenv", "__version__.py")) as f:
    exec (f.read(), about)
if sys.argv[-1] == "publish":
    os.system("python setup.py sdist bdist_wheel upload")
    sys.exit()
required = [
    'pip>=9.0.1',
    'certifi',
    'setuptools>=36.2.1',
    'virtualenv-clone>=0.2.5',
    'virtualenv',
    'pathlib;python_version<"3.4"',
    'requests[security];python_version<"3.0"',
    'ordereddict;python_version<"3.0"',
]


# https://pypi.python.org/pypi/stdeb/0.8.5#quickstart-2-just-tell-me-the-fastest-way-to-make-a-deb
class DebCommand(Command):
    """Support for setup.py deb"""
    description = 'Build and publish the .deb package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'deb_dist'))
        except FileNotFoundError:
            pass
        self.status(u'Creating debian mainfest…')
        os.system(
            'python setup.py --command-packages=stdeb.command sdist_dsc -z artful --package3=pipenv --depends3=python3-virtualenv-clone'
        )
        self.status(u'Building .deb…')
        os.chdir('deb_dist/pipenv-{0}'.format(about['__version__']))
        os.system('dpkg-buildpackage -rfakeroot -uc -us')


class UploadCommand(Command):
    """Support setup.py publish."""
    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
        except FileNotFoundError:
            pass
        self.status('Building Source distribution…')
        os.system('{0} setup.py sdist'.format(sys.executable))
        self.status('Uploading the package to PyPi via Twine…')
        os.system('twine upload dist/*')
        self.status('Pushing git tags…')
        os.system('git tag v{0}'.format(about['__version__']))
        os.system('git push --tags')
        sys.exit()


setup(
    name='pipenv',
    version=about['__version__'],
    description='Python Development Workflow for Humans.',
    long_description=long_description,
    author='Kenneth Reitz',
    author_email='me@kennethreitz.org',
    url='https://github.com/pypa/pipenv',
    packages=find_packages(exclude=['tests']),
    entry_points={
        'console_scripts': [
            'pipenv=pipenv:cli',
            'pewtwo=pipenv.patched.pew.pew:pew',
            'pipenv-resolver=pipenv.resolver:main',
        ]
    },
    install_requires=required,
    extras_require={},
    include_package_data=True,
    license='MIT',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
    cmdclass={'upload': UploadCommand, 'deb': DebCommand},
)



# |~~\'    |~~
# |__/||~~\|--|/~\\  /
# |   ||__/|__|   |\/
#      |
import os
import sys

PIPENV_ROOT = os.path.dirname(os.path.realpath(__file__))
PIPENV_VENDOR = os.sep.join([PIPENV_ROOT, 'vendor'])
PIPENV_PATCHED = os.sep.join([PIPENV_ROOT, 'patched'])
# Inject vendored directory into system path.
sys.path.insert(0, PIPENV_VENDOR)
# Inject patched directory into system path.
sys.path.insert(0, PIPENV_PATCHED)
# Hack to make things work better.
try:
    if 'concurrency' in sys.modules:
        del sys.modules['concurrency']
except Exception:
    pass
from .cli import cli
from . import resolver

if __name__ == '__main__':
    cli()



"""
Partial backport of Python 3.6's weakref module:

    finalize (new in Python 3.4)

Backport modifications are marked with "XXX backport".
"""
__all__ = ["finalize"]

from .weakref import finalize



from prettytoml import tokens
from prettytoml.elements import traversal as t, factory as element_factory
from prettytoml.tokens import py2toml


def table_entries_should_be_uniformly_indented(toml_file_elements):
    """
    Rule: Nth-level table sections should be indented by (N-1)*2 spaces
    """
    elements = toml_file_elements[:]
    for (i, e) in enumerate(elements):
        if t.predicates.table_header(e):
            table = elements[t.find_following(elements, t.predicates.table, i)]
            _do_table_header(e)
            _do_table(table, len(e.names))
    return elements


def _do_table_header(table_header):
    indent_start = 0
    indent_end = next(i for (i, token) in enumerate(table_header.tokens) if token.type != tokens.TYPE_WHITESPACE)

    del table_header.tokens[indent_start:indent_end]
    table_header.tokens.insert(0, py2toml.create_whitespace(' ' * ((len(table_header.names)-1) * 2)))


def _do_table(table_element, table_level):

    elements = table_element.sub_elements

    # Iterator index
    i = float('-inf')

    def first_indent():
        return t.find_following(elements, t.predicates.whitespace, i)

    def next_non_metadata():
        return t.find_following(elements, t.predicates.non_metadata, i)

    def next_newline():
        return t.find_following(elements, t.predicates.newline, next_non_metadata())

    while next_non_metadata() >= 0:
        if first_indent() >= 0:
            del elements[first_indent():next_non_metadata()]

        elements.insert(next_non_metadata(), element_factory.create_whitespace_element((table_level-1)*2))

        i = next_newline()



import os

import pytest


here = os.path.dirname(__file__)
version_file = os.path.join(here, "version.py")

with open(version_file) as f:
    code = compile(f.read(), version_file, 'exec')
    exec(code)

use_class_based_httpbin = pytest.mark.usefixtures("class_based_pypi")
use_class_based_httpbin_secure = pytest.mark.usefixtures("class_based_pypi_secure")



# -*- coding: utf-8 -*-
import sys
import os
import platform
import json


def format_full_version(info):
    version = '{0.major}.{0.minor}.{0.micro}'.format(info)
    kind = info.releaselevel
    if kind != 'final':
        version += kind[0] + str(info.serial)
    return version


# Support for 508's implementation_version.
if hasattr(sys, 'implementation'):
    implementation_version = format_full_version(sys.implementation.version)
else:
    implementation_version = '0'
# Default to cpython for 2.7.
if hasattr(sys, 'implementation'):
    implementation_name = sys.implementation.name
else:
    implementation_name = 'cpython'
lookup = {
    'os_name': os.name,
    'sys_platform': sys.platform,
    'platform_machine': platform.machine(),
    'platform_python_implementation': platform.python_implementation(),
    'platform_release': platform.release(),
    'platform_system': platform.system(),
    'platform_version': platform.version(),
    'python_version': platform.python_version()[:3],
    'python_full_version': platform.python_version(),
    'implementation_name': implementation_name,
    'implementation_version': implementation_version,
}
if __name__ == '__main__':
    print(json.dumps(lookup))



# -*- coding: utf-8 -*-
import logging
import os
import sys

import click
import click_completion
import crayons
import delegator
from click_didyoumean import DYMCommandCollection

from .__version__ import __version__

from .import environments
from .environments import *

# Enable shell completion.
click_completion.init()
CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


class PipenvGroup(click.Group):
    """Custom Group class provides formatted main help"""

    def get_help_option(self, ctx):
        from .import core

        """Override for showing formatted main help via --help and -h options"""
        help_options = self.get_help_option_names(ctx)
        if not help_options or not self.add_help_option:
            return

        def show_help(ctx, param, value):
            if value and not ctx.resilient_parsing:
                if not ctx.invoked_subcommand:
                    # legit main help
                    click.echo(core.format_help(ctx.get_help()))
                else:
                    # legit sub-command help
                    click.echo(ctx.get_help(), color=ctx.color)
                ctx.exit()

        return click.Option(
            help_options,
            is_flag=True,
            is_eager=True,
            expose_value=False,
            callback=show_help,
            help='Show this message and exit.',
        )


def setup_verbose(ctx, param, value):
    if value:
        logging.getLogger('pip').setLevel(logging.INFO)
    return value


@click.group(
    cls=PipenvGroup,
    invoke_without_command=True,
    context_settings=CONTEXT_SETTINGS,
)
@click.option(
    '--where',
    is_flag=True,
    default=False,
    help="Output project home information.",
)
@click.option(
    '--venv',
    is_flag=True,
    default=False,
    help="Output virtualenv information.",
)
@click.option(
    '--py',
    is_flag=True,
    default=False,
    help="Output Python interpreter information.",
)
@click.option(
    '--envs',
    is_flag=True,
    default=False,
    help="Output Environment Variable options.",
)
@click.option(
    '--rm', is_flag=True, default=False, help="Remove the virtualenv."
)
@click.option('--bare', is_flag=True, default=False, help="Minimal output.")
@click.option(
    '--completion',
    is_flag=True,
    default=False,
    help="Output completion (to be eval'd).",
)
@click.option('--man', is_flag=True, default=False, help="Display manpage.")
@click.option(
    '--three/--two',
    is_flag=True,
    default=None,
    help="Use Python 3/2 when creating virtualenv.",
)
@click.option(
    '--python',
    default=False,
    nargs=1,
    help="Specify which version of Python virtualenv should use.",
)
@click.option(
    '--site-packages',
    is_flag=True,
    default=False,
    help="Enable site-packages for the virtualenv.",
)
@click.version_option(
    prog_name=crayons.normal('pipenv', bold=True), version=__version__
)
@click.pass_context
def cli(
    ctx,
    where=False,
    venv=False,
    rm=False,
    bare=False,
    three=False,
    python=False,
    help=False,
    py=False,
    site_packages=False,
    envs=False,
    man=False,
    completion=False,
):
    if completion:  # Handle this ASAP to make shell startup fast.
        if PIPENV_SHELL:
            click.echo(
                click_completion.get_code(
                    shell=PIPENV_SHELL.split(os.sep)[-1], prog_name='pipenv'
                )
            )
        else:
            click.echo(
                'Please ensure that the {0} environment variable '
                'is set.'.format(crayons.normal('SHELL', bold=True)),
                err=True,
            )
            sys.exit(1)
        sys.exit(0)
    from .import core
    if man:
        if core.system_which('man'):
            path = os.sep.join([os.path.dirname(__file__), 'pipenv.1'])
            os.execle(core.system_which('man'), 'man', path, os.environ)
        else:
            click.echo(
                'man does not appear to be available on your system.', err=True
            )
    if envs:
        click.echo(
            'The following environment variables can be set, to do various things:\n'
        )
        for key in environments.__dict__:
            if key.startswith('PIPENV'):
                click.echo('  - {0}'.format(crayons.normal(key, bold=True)))
        click.echo(
            '\nYou can learn more at:\n   {0}'.format(
                crayons.green(
                    'http://docs.pipenv.org/advanced/#configuration-with-environment-variables'
                )
            )
        )
        sys.exit(0)
    core.warn_in_virtualenv()
    if ctx.invoked_subcommand is None:
        # --where was passed...
        if where:
            core.do_where(bare=True)
            sys.exit(0)
        elif py:
            core.do_py()
            sys.exit()
        # --venv was passed...
        elif venv:
            # There is no virtualenv yet.
            if not core.project.virtualenv_exists:
                click.echo(
                    crayons.red(
                        'No virtualenv has been created for this project yet!'
                    ),
                    err=True,
                )
                sys.exit(1)
            else:
                click.echo(core.project.virtualenv_location)
                sys.exit(0)
        # --rm was passed...
        elif rm:
            # Abort if --system (or running in a virtualenv).
            if PIPENV_USE_SYSTEM:
                click.echo(
                    crayons.red(
                        'You are attempting to remove a virtualenv that '
                        'Pipenv did not create. Aborting.'
                    )
                )
                sys.exit(1)
            if core.project.virtualenv_exists:
                loc = core.project.virtualenv_location
                click.echo(
                    crayons.normal(
                        u'{0} ({1})…'.format(
                            crayons.normal('Removing virtualenv', bold=True),
                            crayons.green(loc),
                        )
                    )
                )
                with core.spinner():
                    # Remove the virtualenv.
                    core.cleanup_virtualenv(bare=True)
                sys.exit(0)
            else:
                click.echo(
                    crayons.red(
                        'No virtualenv has been created for this project yet!',
                        bold=True,
                    ),
                    err=True,
                )
                sys.exit(1)
    # --two / --three was passed...
    if (python or three is not None) or site_packages:
        core.ensure_project(
            three=three, python=python, warn=True, site_packages=site_packages
        )
    # Check this again before exiting for empty ``pipenv`` command.
    elif ctx.invoked_subcommand is None:
        # Display help to user, if no commands were passed.
        click.echo(core.format_help(ctx.get_help()))


@click.command(
    short_help="Installs provided packages and adds them to Pipfile, or (if none is given), installs all packages.",
    context_settings=dict(ignore_unknown_options=True, allow_extra_args=True),
)
@click.argument('package_name', default=False)
@click.argument('more_packages', nargs=-1)
@click.option(
    '--dev',
    '-d',
    is_flag=True,
    default=False,
    help="Install package(s) in [dev-packages].",
)
@click.option(
    '--three/--two',
    is_flag=True,
    default=None,
    help="Use Python 3/2 when creating virtualenv.",
)
@click.option(
    '--python',
    default=False,
    nargs=1,
    help="Specify which version of Python virtualenv should use.",
)
@click.option(
    '--system', is_flag=True, default=False, help="System pip management."
)
@click.option(
    '--requirements',
    '-r',
    nargs=1,
    default=False,
    help="Import a requirements.txt file.",
)
@click.option(
    '--code', '-c', nargs=1, default=False, help="Import from codebase."
)
@click.option(
    '--verbose',
    '-v',
    is_flag=True,
    default=False,
    help="Verbose mode.",
    callback=setup_verbose,
)
@click.option(
    '--ignore-pipfile',
    is_flag=True,
    default=False,
    help="Ignore Pipfile when installing, using the Pipfile.lock.",
)
@click.option(
    '--sequential',
    is_flag=True,
    default=False,
    help="Install dependencies one-at-a-time, instead of concurrently.",
)
@click.option(
    '--skip-lock',
    is_flag=True,
    default=False,
    help=u"Ignore locking mechanisms when installing—use the Pipfile, instead.",
)
@click.option(
    '--deploy',
    is_flag=True,
    default=False,
    help=u"Abort if the Pipfile.lock is out–of–date, or Python version is wrong.",
)
@click.option(
    '--pre', is_flag=True, default=False, help=u"Allow pre–releases."
)
@click.option(
    '--keep-outdated',
    is_flag=True,
    default=False,
    help=u"Keep out–dated dependencies from being updated in Pipfile.lock.",
)
@click.option(
    '--selective-upgrade',
    is_flag=True,
    default=False,
    help="Update specified packages.",
)
def install(
    package_name=False,
    more_packages=False,
    dev=False,
    three=False,
    python=False,
    system=False,
    lock=True,
    ignore_pipfile=False,
    skip_lock=False,
    verbose=False,
    requirements=False,
    sequential=False,
    pre=False,
    code=False,
    deploy=False,
    keep_outdated=False,
    selective_upgrade=False,
):
    from .import core

    core.do_install(
        package_name=package_name,
        more_packages=more_packages,
        dev=dev,
        three=three,
        python=python,
        system=system,
        lock=lock,
        ignore_pipfile=ignore_pipfile,
        skip_lock=skip_lock,
        verbose=verbose,
        requirements=requirements,
        sequential=sequential,
        pre=pre,
        code=code,
        deploy=deploy,
        keep_outdated=keep_outdated,
        selective_upgrade=selective_upgrade,
    )


@click.command(
    short_help="Un-installs a provided package and removes it from Pipfile."
)
@click.argument('package_name', default=False)
@click.argument('more_packages', nargs=-1)
@click.option(
    '--three/--two',
    is_flag=True,
    default=None,
    help="Use Python 3/2 when creating virtualenv.",
)
@click.option(
    '--python',
    default=False,
    nargs=1,
    help="Specify which version of Python virtualenv should use.",
)
@click.option(
    '--system', is_flag=True, default=False, help="System pip management."
)
@click.option(
    '--verbose',
    '-v',
    is_flag=True,
    default=False,
    help="Verbose mode.",
    callback=setup_verbose,
)
@click.option('--lock', is_flag=True, default=True, help="Lock afterwards.")
@click.option(
    '--all-dev',
    is_flag=True,
    default=False,
    help="Un-install all package from [dev-packages].",
)
@click.option(
    '--all',
    is_flag=True,
    default=False,
    help="Purge all package(s) from virtualenv. Does not edit Pipfile.",
)
@click.option(
    '--keep-outdated',
    is_flag=True,
    default=False,
    help=u"Keep out–dated dependencies from being updated in Pipfile.lock.",
)
def uninstall(
    package_name=False,
    more_packages=False,
    three=None,
    python=False,
    system=False,
    lock=False,
    all_dev=False,
    all=False,
    verbose=False,
    keep_outdated=False,
):
    from .import core

    core.do_uninstall(
        package_name=package_name,
        more_packages=more_packages,
        three=three,
        python=python,
        system=system,
        lock=lock,
        all_dev=all_dev,
        all=all,
        verbose=verbose,
        keep_outdated=keep_outdated,
    )


@click.command(short_help="Generates Pipfile.lock.")
@click.option(
    '--three/--two',
    is_flag=True,
    default=None,
    help="Use Python 3/2 when creating virtualenv.",
)
@click.option(
    '--python',
    default=False,
    nargs=1,
    help="Specify which version of Python virtualenv should use.",
)
@click.option(
    '--verbose',
    '-v',
    is_flag=True,
    default=False,
    help="Verbose mode.",
    callback=setup_verbose,
)
@click.option(
    '--requirements',
    '-r',
    is_flag=True,
    default=False,
    help="Generate output compatible with requirements.txt.",
)
@click.option(
    '--dev',
    '-d',
    is_flag=True,
    default=False,
    help="Generate output compatible with requirements.txt for the development dependencies.",
)
@click.option(
    '--clear', is_flag=True, default=False, help="Clear the dependency cache."
)
@click.option(
    '--pre', is_flag=True, default=False, help=u"Allow pre–releases."
)
@click.option(
    '--keep-outdated',
    is_flag=True,
    default=False,
    help=u"Keep out–dated dependencies from being updated in Pipfile.lock.",
)
def lock(
    three=None,
    python=False,
    verbose=False,
    requirements=False,
    dev=False,
    clear=False,
    pre=False,
    keep_outdated=False,
):
    from .import core

    # Ensure that virtualenv is available.
    core.ensure_project(three=three, python=python)
    # Load the --pre settings from the Pipfile.
    if not pre:
        pre = core.project.settings.get('pre')
    if requirements:
        core.do_init(dev=dev, requirements=requirements)
    core.do_lock(
        verbose=verbose, clear=clear, pre=pre, keep_outdated=keep_outdated
    )


@click.command(
    short_help="Spawns a shell within the virtualenv.",
    context_settings=dict(ignore_unknown_options=True, allow_extra_args=True),
)
@click.option(
    '--three/--two',
    is_flag=True,
    default=None,
    help="Use Python 3/2 when creating virtualenv.",
)
@click.option(
    '--python',
    default=False,
    nargs=1,
    help="Specify which version of Python virtualenv should use.",
)
@click.option(
    '--fancy',
    is_flag=True,
    default=False,
    help="Run in shell in fancy mode (for elegantly configured shells).",
)
@click.option(
    '--anyway',
    is_flag=True,
    default=False,
    help="Always spawn a subshell, even if one is already spawned.",
)
@click.argument('shell_args', nargs=-1)
def shell(
    three=None, python=False, fancy=False, shell_args=None, anyway=False
):
    from .import core

    # Prevent user from activating nested environments.
    if 'PIPENV_ACTIVE' in os.environ:
        # If PIPENV_ACTIVE is set, VIRTUAL_ENV should always be set too.
        venv_name = os.environ.get(
            'VIRTUAL_ENV', 'UNKNOWN_VIRTUAL_ENVIRONMENT'
        )
        if not anyway:
            click.echo(
                '{0} {1} {2}\nNo action taken to avoid nested environments.'.format(
                    crayons.normal('Shell for'),
                    crayons.green(venv_name, bold=True),
                    crayons.normal('already activated.', bold=True),
                ),
                err=True,
            )
            sys.exit(1)
    # Load .env file.
    core.load_dot_env()
    # Use fancy mode for Windows.
    if os.name == 'nt':
        fancy = True
    core.do_shell(
        three=three, python=python, fancy=fancy, shell_args=shell_args
    )


@click.command(
    add_help_option=False,
    short_help="Spawns a command installed into the virtualenv.",
    context_settings=dict(
        ignore_unknown_options=True,
        allow_interspersed_args=False,
        allow_extra_args=True,
    ),
)
@click.argument('command')
@click.argument('args', nargs=-1)
@click.option(
    '--three/--two',
    is_flag=True,
    default=None,
    help="Use Python 3/2 when creating virtualenv.",
)
@click.option(
    '--python',
    default=False,
    nargs=1,
    help="Specify which version of Python virtualenv should use.",
)
def run(command, args, three=None, python=False):
    from .import core

    core.do_run(command=command, args=args, three=three, python=python)


@click.command(
    short_help="Checks for security vulnerabilities and against PEP 508 markers provided in Pipfile.",
    context_settings=dict(ignore_unknown_options=True, allow_extra_args=True),
)
@click.option(
    '--three/--two',
    is_flag=True,
    default=None,
    help="Use Python 3/2 when creating virtualenv.",
)
@click.option(
    '--python',
    default=False,
    nargs=1,
    help="Specify which version of Python virtualenv should use.",
)
@click.option(
    '--system', is_flag=True, default=False, help="Use system Python."
)
@click.option(
    '--unused',
    nargs=1,
    default=False,
    help="Given a code path, show potentially unused dependencies.",
)
@click.argument('args', nargs=-1)
def check(
    three=None,
    python=False,
    system=False,
    unused=False,
    style=False,
    args=None,
):
    from .import core

    core.do_check(
        three=three, python=python, system=system, unused=unused, args=args
    )


@click.command(short_help="Runs lock, then sync.")
@click.argument('more_packages', nargs=-1)
@click.option(
    '--three/--two',
    is_flag=True,
    default=None,
    help="Use Python 3/2 when creating virtualenv.",
)
@click.option(
    '--python',
    default=False,
    nargs=1,
    help="Specify which version of Python virtualenv should use.",
)
@click.option(
    '--verbose',
    '-v',
    is_flag=True,
    default=False,
    help="Verbose mode.",
    callback=setup_verbose,
)
@click.option(
    '--dev',
    '-d',
    is_flag=True,
    default=False,
    help="Install package(s) in [dev-packages].",
)
@click.option(
    '--clear', is_flag=True, default=False, help="Clear the dependency cache."
)
@click.option('--bare', is_flag=True, default=False, help="Minimal output.")
@click.option(
    '--pre', is_flag=True, default=False, help=u"Allow pre–releases."
)
@click.option(
    '--keep-outdated',
    is_flag=True,
    default=False,
    help=u"Keep out–dated dependencies from being updated in Pipfile.lock.",
)
@click.option(
    '--sequential',
    is_flag=True,
    default=False,
    help="Install dependencies one-at-a-time, instead of concurrently.",
)
@click.option(
    '--outdated',
    is_flag=True,
    default=False,
    help=u"List out–of–date dependencies.",
)
@click.option(
    '--dry-run',
    is_flag=True,
    default=None,
    help=u"List out–of–date dependencies.",
)
@click.argument('package', default=False)
@click.pass_context
def update(
    ctx,
    three=None,
    python=False,
    system=False,
    verbose=False,
    clear=False,
    keep_outdated=False,
    pre=False,
    dev=False,
    bare=False,
    sequential=False,
    package=None,
    dry_run=None,
    outdated=False,
    more_packages=None,
):
    from .import core

    core.ensure_project(three=three, python=python, warn=True)
    if not outdated:
        outdated = bool(dry_run)
    if outdated:
        core.do_outdated()
    if not package:
        click.echo(
            '{0} {1} {2} {3}{4}'.format(
                crayons.white('Running', bold=True),
                crayons.red('$ pipenv lock', bold=True),
                crayons.white('then', bold=True),
                crayons.red('$ pipenv sync', bold=True),
                crayons.white('.', bold=True),
            )
        )
        # Load the --pre settings from the Pipfile.
        if not pre:
            pre = core.project.settings.get('pre')
        core.do_lock(
            verbose=verbose, clear=clear, pre=pre, keep_outdated=keep_outdated
        )
        core.do_sync(
            ctx=ctx,
            install=install,
            dev=dev,
            three=three,
            python=python,
            bare=bare,
            dont_upgrade=False,
            user=False,
            verbose=verbose,
            clear=clear,
            unused=False,
            sequential=sequential,
        )
    else:
        for package in ([package] + list(more_packages) or []):
            if package not in core.project.all_packages:
                click.echo(
                    '{0}: {1} was not found in your Pipfile! Aborting.'
                    ''.format(
                        crayons.red('Warning', bold=True),
                        crayons.green(package, bold=True),
                    ),
                    err=True,
                )
                sys.exit(1)
        core.ensure_lockfile(keep_outdated=core.project.lockfile_exists)
        # Install the dependencies.
        core.do_install(
            package_name=package,
            more_packages=more_packages,
            dev=dev,
            three=three,
            python=python,
            system=system,
            lock=True,
            ignore_pipfile=False,
            skip_lock=False,
            verbose=verbose,
            requirements=False,
            sequential=sequential,
            pre=pre,
            code=False,
            deploy=False,
            keep_outdated=True,
            selective_upgrade=True,
        )


@click.command(
    short_help=u"Displays currently–installed dependency graph information."
)
@click.option('--bare', is_flag=True, default=False, help="Minimal output.")
@click.option('--json', is_flag=True, default=False, help="Output JSON.")
@click.option(
    '--reverse', is_flag=True, default=False, help="Reversed dependency graph."
)
def graph(bare=False, json=False, reverse=False):
    from .import core

    core.do_graph(bare=bare, json=json, reverse=reverse)


@click.command(short_help="View a given module in your editor.", name="open")
@click.option(
    '--three/--two',
    is_flag=True,
    default=None,
    help="Use Python 3/2 when creating virtualenv.",
)
@click.option(
    '--python',
    default=False,
    nargs=1,
    help="Specify which version of Python virtualenv should use.",
)
@click.argument('module', nargs=1)
def run_open(module, three=None, python=None):
    from .import core

    # Ensure that virtualenv is available.
    core.ensure_project(three=three, python=python, validate=False)
    c = delegator.run(
        '{0} -c "import {1}; print({1}.__file__);"'.format(
            core.which('python'), module
        )
    )
    try:
        assert c.return_code == 0
    except AssertionError:
        click.echo(crayons.red('Module not found!'))
        sys.exit(1)
    if '__init__.py' in c.out:
        p = os.path.dirname(c.out.strip().rstrip('cdo'))
    else:
        p = c.out.strip().rstrip('cdo')
    click.echo(
        crayons.normal('Opening {0!r} in your EDITOR.'.format(p), bold=True)
    )
    click.edit(filename=p)
    sys.exit(0)


@click.command(short_help="Installs all packages specified in Pipfile.lock.")
@click.option(
    '--verbose',
    '-v',
    is_flag=True,
    default=False,
    help="Verbose mode.",
    callback=setup_verbose,
)
@click.option(
    '--dev',
    '-d',
    is_flag=True,
    default=False,
    help="Additionally install package(s) in [dev-packages].",
)
@click.option(
    '--three/--two',
    is_flag=True,
    default=None,
    help="Use Python 3/2 when creating virtualenv.",
)
@click.option(
    '--python',
    default=False,
    nargs=1,
    help="Specify which version of Python virtualenv should use.",
)
@click.option('--bare', is_flag=True, default=False, help="Minimal output.")
@click.option(
    '--clear', is_flag=True, default=False, help="Clear the dependency cache."
)
@click.option(
    '--sequential',
    is_flag=True,
    default=False,
    help="Install dependencies one-at-a-time, instead of concurrently.",
)
@click.pass_context
def sync(
    ctx,
    dev=False,
    three=None,
    python=None,
    bare=False,
    dont_upgrade=False,
    user=False,
    verbose=False,
    clear=False,
    unused=False,
    package_name=None,
    sequential=False,
):
    from .import core

    core.do_sync(
        ctx=ctx,
        install=install,
        dev=dev,
        three=three,
        python=python,
        bare=bare,
        dont_upgrade=dont_upgrade,
        user=user,
        verbose=verbose,
        clear=clear,
        unused=unused,
        sequential=sequential,
    )


@click.command(
    short_help="Uninstalls all packages not specified in Pipfile.lock."
)
@click.option(
    '--verbose',
    '-v',
    is_flag=True,
    default=False,
    help="Verbose mode.",
    callback=setup_verbose,
)
@click.option(
    '--three/--two',
    is_flag=True,
    default=None,
    help="Use Python 3/2 when creating virtualenv.",
)
@click.option(
    '--python',
    default=False,
    nargs=1,
    help="Specify which version of Python virtualenv should use.",
)
@click.option(
    '--dry-run',
    is_flag=True,
    default=False,
    help="Just output unneeded packages.",
)
@click.pass_context
def clean(
    ctx,
    three=None,
    python=None,
    dry_run=False,
    bare=False,
    user=False,
    verbose=False,
):
    from .import core

    core.do_clean(
        ctx=ctx, three=three, python=python, dry_run=dry_run, verbose=verbose
    )


# Install click commands.
cli.add_command(graph)
cli.add_command(install)
cli.add_command(uninstall)
cli.add_command(sync)
cli.add_command(lock)
cli.add_command(check)
cli.add_command(clean)
cli.add_command(shell)
cli.add_command(run)
cli.add_command(update)
cli.add_command(run_open)
# Only invoke the "did you mean" when an argument wasn't passed (it breaks those).
if '-' not in ''.join(sys.argv) and len(sys.argv) > 1:
    cli = DYMCommandCollection(sources=[cli])
if __name__ == '__main__':
    cli()



from __future__ import absolute_import

import logging
import operator
import os
import tempfile
import shutil
import warnings
try:
    import wheel
except ImportError:
    wheel = None

from pip9.req import RequirementSet
from pip9.basecommand import RequirementCommand
from pip9.locations import virtualenv_no_global, distutils_scheme
from pip9.exceptions import (
    InstallationError, CommandError, PreviousBuildDirError,
)
from pip9 import cmdoptions
from pip9.utils import ensure_dir, get_installed_version
from pip9.utils.build import BuildDirectory
from pip9.utils.deprecation import RemovedInPip10Warning
from pip9.utils.filesystem import check_path_owner
from pip9.wheel import WheelCache, WheelBuilder


logger = logging.getLogger(__name__)


class InstallCommand(RequirementCommand):
    """
    Install packages from:

    - PyPI (and other indexes) using requirement specifiers.
    - VCS project urls.
    - Local project directories.
    - Local or remote source archives.

    pip also supports installing from "requirements files", which provide
    an easy way to specify a whole environment to be installed.
    """
    name = 'install'

    usage = """
      %prog [options] <requirement specifier> [package-index-options] ...
      %prog [options] -r <requirements file> [package-index-options] ...
      %prog [options] [-e] <vcs project url> ...
      %prog [options] [-e] <local project path> ...
      %prog [options] <archive url/path> ..."""

    summary = 'Install packages.'

    def __init__(self, *args, **kw):
        super(InstallCommand, self).__init__(*args, **kw)

        cmd_opts = self.cmd_opts

        cmd_opts.add_option(cmdoptions.constraints())
        cmd_opts.add_option(cmdoptions.editable())
        cmd_opts.add_option(cmdoptions.requirements())
        cmd_opts.add_option(cmdoptions.build_dir())

        cmd_opts.add_option(
            '-t', '--target',
            dest='target_dir',
            metavar='dir',
            default=None,
            help='Install packages into <dir>. '
                 'By default this will not replace existing files/folders in '
                 '<dir>. Use --upgrade to replace existing packages in <dir> '
                 'with new versions.'
        )

        cmd_opts.add_option(
            '-d', '--download', '--download-dir', '--download-directory',
            dest='download_dir',
            metavar='dir',
            default=None,
            help=("Download packages into <dir> instead of installing them, "
                  "regardless of what's already installed."),
        )

        cmd_opts.add_option(cmdoptions.src())

        cmd_opts.add_option(
            '-U', '--upgrade',
            dest='upgrade',
            action='store_true',
            help='Upgrade all specified packages to the newest available '
                 'version. The handling of dependencies depends on the '
                 'upgrade-strategy used.'
        )

        cmd_opts.add_option(
            '--upgrade-strategy',
            dest='upgrade_strategy',
            default='eager',
            choices=['only-if-needed', 'eager'],
            help='Determines how dependency upgrading should be handled. '
                 '"eager" - dependencies are upgraded regardless of '
                 'whether the currently installed version satisfies the '
                 'requirements of the upgraded package(s). '
                 '"only-if-needed" -  are upgraded only when they do not '
                 'satisfy the requirements of the upgraded package(s).'
        )

        cmd_opts.add_option(
            '--force-reinstall',
            dest='force_reinstall',
            action='store_true',
            help='When upgrading, reinstall all packages even if they are '
                 'already up-to-date.')

        cmd_opts.add_option(
            '-I', '--ignore-installed',
            dest='ignore_installed',
            action='store_true',
            help='Ignore the installed packages (reinstalling instead).')

        cmd_opts.add_option(cmdoptions.ignore_requires_python())
        cmd_opts.add_option(cmdoptions.no_deps())

        cmd_opts.add_option(cmdoptions.install_options())
        cmd_opts.add_option(cmdoptions.global_options())

        cmd_opts.add_option(
            '--user',
            dest='use_user_site',
            action='store_true',
            help="Install to the Python user install directory for your "
                 "platform. Typically ~/.local/, or %APPDATA%\Python on "
                 "Windows. (See the Python documentation for site.USER_BASE "
                 "for full details.)")

        cmd_opts.add_option(
            '--egg',
            dest='as_egg',
            action='store_true',
            help="Install packages as eggs, not 'flat', like pip normally "
                 "does. This option is not about installing *from* eggs. "
                 "(WARNING: Because this option overrides pip's normal install"
                 " logic, requirements files may not behave as expected.)")

        cmd_opts.add_option(
            '--root',
            dest='root_path',
            metavar='dir',
            default=None,
            help="Install everything relative to this alternate root "
                 "directory.")

        cmd_opts.add_option(
            '--prefix',
            dest='prefix_path',
            metavar='dir',
            default=None,
            help="Installation prefix where lib, bin and other top-level "
                 "folders are placed")

        cmd_opts.add_option(
            "--compile",
            action="store_true",
            dest="compile",
            default=True,
            help="Compile py files to pyc",
        )

        cmd_opts.add_option(
            "--no-compile",
            action="store_false",
            dest="compile",
            help="Do not compile py files to pyc",
        )

        cmd_opts.add_option(cmdoptions.use_wheel())
        cmd_opts.add_option(cmdoptions.no_use_wheel())
        cmd_opts.add_option(cmdoptions.no_binary())
        cmd_opts.add_option(cmdoptions.only_binary())
        cmd_opts.add_option(cmdoptions.pre())
        cmd_opts.add_option(cmdoptions.no_clean())
        cmd_opts.add_option(cmdoptions.require_hashes())

        index_opts = cmdoptions.make_option_group(
            cmdoptions.index_group,
            self.parser,
        )

        self.parser.insert_option_group(0, index_opts)
        self.parser.insert_option_group(0, cmd_opts)

    def run(self, options, args):
        cmdoptions.resolve_wheel_no_use_binary(options)
        cmdoptions.check_install_build_global(options)

        if options.as_egg:
            warnings.warn(
                "--egg has been deprecated and will be removed in the future. "
                "This flag is mutually exclusive with large parts of pip, and "
                "actually using it invalidates pip's ability to manage the "
                "installation process.",
                RemovedInPip10Warning,
            )

        if options.allow_external:
            warnings.warn(
                "--allow-external has been deprecated and will be removed in "
                "the future. Due to changes in the repository protocol, it no "
                "longer has any effect.",
                RemovedInPip10Warning,
            )

        if options.allow_all_external:
            warnings.warn(
                "--allow-all-external has been deprecated and will be removed "
                "in the future. Due to changes in the repository protocol, it "
                "no longer has any effect.",
                RemovedInPip10Warning,
            )

        if options.allow_unverified:
            warnings.warn(
                "--allow-unverified has been deprecated and will be removed "
                "in the future. Due to changes in the repository protocol, it "
                "no longer has any effect.",
                RemovedInPip10Warning,
            )

        if options.download_dir:
            warnings.warn(
                "pip install --download has been deprecated and will be "
                "removed in the future. Pip now has a download command that "
                "should be used instead.",
                RemovedInPip10Warning,
            )
            options.ignore_installed = True

        if options.build_dir:
            options.build_dir = os.path.abspath(options.build_dir)

        options.src_dir = os.path.abspath(options.src_dir)
        install_options = options.install_options or []
        if options.use_user_site:
            if options.prefix_path:
                raise CommandError(
                    "Can not combine '--user' and '--prefix' as they imply "
                    "different installation locations"
                )
            if virtualenv_no_global():
                raise InstallationError(
                    "Can not perform a '--user' install. User site-packages "
                    "are not visible in this virtualenv."
                )
            install_options.append('--user')
            install_options.append('--prefix=')

        temp_target_dir = None
        if options.target_dir:
            options.ignore_installed = True
            temp_target_dir = tempfile.mkdtemp()
            options.target_dir = os.path.abspath(options.target_dir)
            if (os.path.exists(options.target_dir) and not
                    os.path.isdir(options.target_dir)):
                raise CommandError(
                    "Target path exists but is not a directory, will not "
                    "continue."
                )
            install_options.append('--home=' + temp_target_dir)

        global_options = options.global_options or []

        with self._build_session(options) as session:

            finder = self._build_package_finder(options, session)
            build_delete = (not (options.no_clean or options.build_dir))
            wheel_cache = WheelCache(options.cache_dir, options.format_control)
            if options.cache_dir and not check_path_owner(options.cache_dir):
                logger.warning(
                    "The directory '%s' or its parent directory is not owned "
                    "by the current user and caching wheels has been "
                    "disabled. check the permissions and owner of that "
                    "directory. If executing pip with sudo, you may want "
                    "sudo's -H flag.",
                    options.cache_dir,
                )
                options.cache_dir = None

            with BuildDirectory(options.build_dir,
                                delete=build_delete) as build_dir:
                requirement_set = RequirementSet(
                    build_dir=build_dir,
                    src_dir=options.src_dir,
                    download_dir=options.download_dir,
                    upgrade=options.upgrade,
                    upgrade_strategy=options.upgrade_strategy,
                    as_egg=options.as_egg,
                    ignore_installed=options.ignore_installed,
                    ignore_dependencies=options.ignore_dependencies,
                    ignore_requires_python=options.ignore_requires_python,
                    force_reinstall=options.force_reinstall,
                    use_user_site=options.use_user_site,
                    target_dir=temp_target_dir,
                    session=session,
                    pycompile=options.compile,
                    isolated=options.isolated_mode,
                    wheel_cache=wheel_cache,
                    require_hashes=options.require_hashes,
                )

                self.populate_requirement_set(
                    requirement_set, args, options, finder, session, self.name,
                    wheel_cache
                )

                if not requirement_set.has_requirements:
                    return

                try:
                    if (options.download_dir or not wheel or not
                            options.cache_dir):
                        # on -d don't do complex things like building
                        # wheels, and don't try to build wheels when wheel is
                        # not installed.
                        requirement_set.prepare_files(finder)
                    else:
                        # build wheels before install.
                        wb = WheelBuilder(
                            requirement_set,
                            finder,
                            build_options=[],
                            global_options=[],
                        )
                        # Ignore the result: a failed wheel will be
                        # installed from the sdist/vcs whatever.
                        wb.build(autobuilding=True)

                    if not options.download_dir:
                        requirement_set.install(
                            install_options,
                            global_options,
                            root=options.root_path,
                            prefix=options.prefix_path,
                        )

                        possible_lib_locations = get_lib_location_guesses(
                            user=options.use_user_site,
                            home=temp_target_dir,
                            root=options.root_path,
                            prefix=options.prefix_path,
                            isolated=options.isolated_mode,
                        )
                        reqs = sorted(
                            requirement_set.successfully_installed,
                            key=operator.attrgetter('name'))
                        items = []
                        for req in reqs:
                            item = req.name
                            try:
                                installed_version = get_installed_version(
                                    req.name, possible_lib_locations
                                )
                                if installed_version:
                                    item += '-' + installed_version
                            except Exception:
                                pass
                            items.append(item)
                        installed = ' '.join(items)
                        if installed:
                            logger.info('Successfully installed %s', installed)
                    else:
                        downloaded = ' '.join([
                            req.name
                            for req in requirement_set.successfully_downloaded
                        ])
                        if downloaded:
                            logger.info(
                                'Successfully downloaded %s', downloaded
                            )
                except PreviousBuildDirError:
                    options.no_clean = True
                    raise
                finally:
                    # Clean up
                    if not options.no_clean:
                        requirement_set.cleanup_files()

        if options.target_dir:
            ensure_dir(options.target_dir)

            # Checking both purelib and platlib directories for installed
            # packages to be moved to target directory
            lib_dir_list = []

            purelib_dir = distutils_scheme('', home=temp_target_dir)['purelib']
            platlib_dir = distutils_scheme('', home=temp_target_dir)['platlib']

            if os.path.exists(purelib_dir):
                lib_dir_list.append(purelib_dir)
            if os.path.exists(platlib_dir) and platlib_dir != purelib_dir:
                lib_dir_list.append(platlib_dir)

            for lib_dir in lib_dir_list:
                for item in os.listdir(lib_dir):
                    target_item_dir = os.path.join(options.target_dir, item)
                    if os.path.exists(target_item_dir):
                        if not options.upgrade:
                            logger.warning(
                                'Target directory %s already exists. Specify '
                                '--upgrade to force replacement.',
                                target_item_dir
                            )
                            continue
                        if os.path.islink(target_item_dir):
                            logger.warning(
                                'Target directory %s already exists and is '
                                'a link. Pip will not automatically replace '
                                'links, please remove if replacement is '
                                'desired.',
                                target_item_dir
                            )
                            continue
                        if os.path.isdir(target_item_dir):
                            shutil.rmtree(target_item_dir)
                        else:
                            os.remove(target_item_dir)

                    shutil.move(
                        os.path.join(lib_dir, item),
                        target_item_dir
                    )
            shutil.rmtree(temp_target_dir)
        return requirement_set


def get_lib_location_guesses(*args, **kwargs):
    scheme = distutils_scheme('', *args, **kwargs)
    return [scheme['purelib'], scheme['platlib']]



# coding: utf-8
from __future__ import absolute_import, division, print_function

import os as _os
import sys as _sys
import warnings as _warnings
from tempfile import mkdtemp


class TemporaryDirectory(object):
    """Create and return a temporary directory.  This has the same
    behavior as mkdtemp but can be used as a context manager.  For
    example:

        with TemporaryDirectory() as tmpdir:
            ...

    Upon exiting the context, the directory and everything contained
    in it are removed.
    """

    def __init__(self, suffix="", prefix="tmp", dir=None):
        self._closed = False
        self.name = None  # Handle mkdtemp raising an exception
        self.name = mkdtemp(suffix, prefix, dir)

    def __repr__(self):
        return "<{} {!r}>".format(self.__class__.__name__, self.name)

    def __enter__(self):
        return self.name

    def cleanup(self):
        if self.name and not self._closed:
            try:
                self._rmtree(self.name)
            except (TypeError, AttributeError) as ex:
                # Issue #10188: Emit a warning on stderr
                # if the directory could not be cleaned
                # up due to missing globals
                if "None" not in str(ex):
                    raise
                print("ERROR: {!r} while cleaning up {!r}".format(ex, self,),
                      file=_sys.stderr)
                return
            self._closed = True

    def __exit__(self, exc, value, tb):
        self.cleanup()

    def __del__(self):
        # Issue a ResourceWarning if implicit cleanup needed
        self.cleanup()

    # XXX (ncoghlan): The following code attempts to make
    # this class tolerant of the module nulling out process
    # that happens during CPython interpreter shutdown
    # Alas, it doesn't actually manage it. See issue #10188
    _listdir = staticmethod(_os.listdir)
    _path_join = staticmethod(_os.path.join)
    _isdir = staticmethod(_os.path.isdir)
    _islink = staticmethod(_os.path.islink)
    _remove = staticmethod(_os.remove)
    _rmdir = staticmethod(_os.rmdir)
    _warn = _warnings.warn

    def _rmtree(self, path):
        # Essentially a stripped down version of shutil.rmtree.  We can't
        # use globals because they may be None'ed out at shutdown.
        for name in self._listdir(path):
            fullname = self._path_join(path, name)
            try:
                isdir = self._isdir(fullname) and not self._islink(fullname)
            except OSError:
                isdir = False
            if isdir:
                self._rmtree(fullname)
            else:
                try:
                    self._remove(fullname)
                except OSError:
                    pass
        try:
            self._rmdir(path)
        except OSError:
            pass



# -*- coding: utf-8 -*-
import errno
import os
import hashlib
import tempfile
import sys
import shutil
import logging
import errno
import click
import crayons
import delegator
import parse
import requests
import six
import stat
import warnings

try:
    from weakref import finalize
except ImportError:
    try:
        from backports.weakref import finalize
    except ImportError:
        pass
from time import time

logging.basicConfig(level=logging.ERROR)
try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse
try:
    from pathlib import Path
except ImportError:
    try:
        from pathlib2 import Path
    except ImportError:
        pass
from distutils.spawn import find_executable
from contextlib import contextmanager
from pipenv.patched.piptools.resolver import Resolver
from pipenv.patched.piptools.repositories.pypi import PyPIRepository
from pipenv.patched.piptools.scripts.compile import get_pip_command
from pipenv.patched.piptools import logging as piptools_logging
from pipenv.patched.piptools.exceptions import NoCandidateFound
from pip9.download import is_archive_file
from pip9.exceptions import DistributionNotFound
from pip9.index import Link
from pip9._vendor.requests.exceptions import HTTPError, ConnectionError

from .pep508checker import lookup
from .environments import SESSION_IS_INTERACTIVE, PIPENV_MAX_ROUNDS, PIPENV_CACHE_DIR

if six.PY2:

    class ResourceWarning(Warning):
        pass


specifiers = [k for k in lookup.keys()]
# List of version control systems we support.
VCS_LIST = ('git', 'svn', 'hg', 'bzr')
SCHEME_LIST = ('http://', 'https://', 'ftp://', 'file://')
requests = requests.Session()


def get_requirement(dep):
    from pip9.req.req_install import _strip_extras
    import requirements

    """Pre-clean requirement strings passed to the requirements parser.

    Ensures that we can accept both local and relative paths, file and VCS URIs,
    remote URIs, and package names, and that we pass only valid requirement strings
    to the requirements parser. Performs necessary modifications to requirements
    object if the user input was a local relative path.

    :param str dep: A requirement line
    :returns: :class:`requirements.Requirement` object
    """
    path = None
    uri = None
    cleaned_uri = None
    editable = False
    dep_link = None
    # check for editable dep / vcs dep
    if dep.startswith('-e '):
        editable = True
        # Use the user supplied path as the written dependency
        dep = dep.split(' ', 1)[1]
    # Split out markers if they are present - similar to how pip does it
    # See pip9.req.req_install.InstallRequirement.from_line
    if not any(dep.startswith(uri_prefix) for uri_prefix in SCHEME_LIST):
        marker_sep = ';'
    else:
        marker_sep = '; '
    if marker_sep in dep:
        dep, markers = dep.split(marker_sep, 1)
        markers = markers.strip()
        if not markers:
            markers = None
    else:
        markers = None
    # Strip extras from the requirement so we can make a properly parseable req
    dep, extras = _strip_extras(dep)
    # Only operate on local, existing, non-URI formatted paths which are installable
    if is_installable_file(dep):
        dep_path = Path(dep)
        dep_link = Link(dep_path.absolute().as_uri())
        if dep_path.is_absolute() or dep_path.as_posix() == '.':
            path = dep_path.as_posix()
        else:
            path = get_converted_relative_path(dep)
        dep = dep_link.egg_fragment if dep_link.egg_fragment else dep_link.url_without_fragment
    elif is_vcs(dep):
        # Generate a Link object for parsing egg fragments
        dep_link = Link(dep)
        # Save the original path to store in the pipfile
        uri = dep_link.url
        # Construct the requirement using proper git+ssh:// replaced uris or names if available
        cleaned_uri = clean_git_uri(dep)
        dep = cleaned_uri
    if editable:
        dep = '-e {0}'.format(dep)
    req = [r for r in requirements.parse(dep)][0]
    # if all we built was the requirement name and still need everything else
    if req.name and not any([req.uri, req.path]):
        if dep_link:
            if dep_link.scheme.startswith('file') and path and not req.path:
                req.path = path
                req.local_file = True
                req.uri = None
            else:
                req.uri = dep_link.url_without_fragment
    # If the result is a local file with a URI and we have a local path, unset the URI
    # and set the path instead -- note that local files may have 'path' set by accident
    elif req.local_file and path and not req.vcs:
        req.path = path
        req.uri = None
    elif req.vcs and req.uri and cleaned_uri and cleaned_uri != uri:
        req.uri = strip_ssh_from_git_uri(req.uri)
        req.line = strip_ssh_from_git_uri(req.line)
    req.editable = editable
    if markers:
        req.markers = markers
    if extras:
        # Bizarrely this is also what pip does...
        req.extras = [
            r for r in requirements.parse('fakepkg{0}'.format(extras))
        ][
            0
        ].extras
    return req


def cleanup_toml(tml):
    toml = tml.split('\n')
    new_toml = []
    # Remove all empty lines from TOML.
    for line in toml:
        if line.strip():
            new_toml.append(line)
    toml = '\n'.join(new_toml)
    new_toml = []
    # Add newlines between TOML sections.
    for i, line in enumerate(toml.split('\n')):
        # Skip the first line.
        if line.startswith('['):
            if i > 0:
                # Insert a newline before the heading.
                new_toml.append('')
        new_toml.append(line)
    # adding new line at the end of the TOML file
    new_toml.append('')
    toml = '\n'.join(new_toml)
    return toml


def python_version(path_to_python):
    if not path_to_python:
        return None

    try:
        c = delegator.run([path_to_python, '--version'], block=False)
    except Exception:
        return None

    output = c.out.strip() or c.err.strip()

    @parse.with_pattern(r'.*')
    def allow_empty(text):
        return text

    TEMPLATE = 'Python {}.{}.{:d}{:AllowEmpty}'
    parsed = parse.parse(TEMPLATE, output, dict(AllowEmpty=allow_empty))
    if parsed:
        parsed = parsed.fixed
    else:
        return None

    return u"{v[0]}.{v[1]}.{v[2]}".format(v=parsed)


def escape_grouped_arguments(s):
    """Prepares a string for the shell (on Windows too!)

    Only for use on grouped arguments (passed as a string to Popen)
    """
    if s is None:
        return None

    # Additional escaping for windows paths
    if os.name == 'nt':
        s = "{}".format(s.replace("\\", "\\\\"))
    return '"' + s.replace("'", "'\\''") + '"'


def clean_pkg_version(version):
    """Uses pip to prepare a package version string, from our internal version."""
    return six.u(pep440_version(str(version).replace('==', '')))


class HackedPythonVersion(object):
    """A Beautiful hack, which allows us to tell pip which version of Python we're using."""

    def __init__(self, python_version, python_path):
        self.python_version = python_version
        self.python_path = python_path

    def __enter__(self):
        os.environ['PIP_PYTHON_VERSION'] = str(self.python_version)
        os.environ['PIP_PYTHON_PATH'] = str(self.python_path)

    def __exit__(self, *args):
        # Restore original Python version information.
        del os.environ['PIP_PYTHON_VERSION']


def prepare_pip_source_args(sources, pip_args=None):
    if pip_args is None:
        pip_args = []
    if sources:
        # Add the source to pip9.
        pip_args.extend(['-i', sources[0]['url']])
        # Trust the host if it's not verified.
        if not sources[0].get('verify_ssl', True):
            pip_args.extend(
                [
                    '--trusted-host',
                    urlparse(sources[0]['url']).netloc.split(':')[0],
                ]
            )
        # Add additional sources as extra indexes.
        if len(sources) > 1:
            for source in sources[1:]:
                pip_args.extend(['--extra-index-url', source['url']])
                # Trust the host if it's not verified.
                if not source.get('verify_ssl', True):
                    pip_args.extend(
                        [
                            '--trusted-host',
                            urlparse(source['url']).hostname,
                        ]
                    )
    return pip_args


def actually_resolve_reps(
    deps, index_lookup, markers_lookup, project, sources, verbose, clear, pre
):
    from pip9 import basecommand, req
    from pip9._vendor import requests as pip_requests

    class PipCommand(basecommand.Command):
        """Needed for pip-tools."""
        name = 'PipCommand'

    constraints = []
    req_dir = tempfile.mkdtemp(prefix='pipenv-', suffix='-requirements')
    for dep in deps:
        if dep:
            if dep.startswith('-e '):
                constraint = req.InstallRequirement.from_editable(
                    dep[len('-e '):]
                )
            else:
                fd, t = tempfile.mkstemp(
                    prefix='pipenv-', suffix='-requirement.txt', dir=req_dir
                )
                with os.fdopen(fd, 'w') as f:
                    f.write(dep)
                constraint = [
                    c for c in req.parse_requirements(t, session=pip_requests)
                ][
                    0
                ]
            # extra_constraints = []
            if ' -i ' in dep:
                index_lookup[constraint.name] = project.get_source(
                    url=dep.split(' -i ')[1]
                ).get(
                    'name'
                )
            if constraint.markers:
                markers_lookup[constraint.name] = str(
                    constraint.markers
                ).replace(
                    '"', "'"
                )
            constraints.append(constraint)
    rmtree(req_dir)
    pip_command = get_pip_command()
    pip_args = []
    if sources:
        pip_args = prepare_pip_source_args(sources, pip_args)
    if verbose:
        print('Using pip: {0}'.format(' '.join(pip_args)))
    pip_options, _ = pip_command.parse_args(pip_args)
    session = pip_command._build_session(pip_options)
    pypi = PyPIRepository(
        pip_options=pip_options, use_json=False, session=session
    )
    if verbose:
        logging.log.verbose = True
        piptools_logging.log.verbose = True
    resolved_tree = set()
    resolver = Resolver(
        constraints=constraints,
        repository=pypi,
        clear_caches=clear,
        prereleases=pre,
    )
    # pre-resolve instead of iterating to avoid asking pypi for hashes of editable packages
    try:
        resolved_tree.update(resolver.resolve(max_rounds=PIPENV_MAX_ROUNDS))
    except (NoCandidateFound, DistributionNotFound, HTTPError) as e:
        click.echo(
            '{0}: Your dependencies could not be resolved. You likely have a mismatch in your sub-dependencies.\n  '
            'You can use {1} to bypass this mechanism, then run {2} to inspect the situation.'
            ''.format(
                crayons.red('Warning', bold=True),
                crayons.red('$ pipenv install --skip-lock'),
                crayons.red('$ pipenv graph'),
            ),
            err=True,
        )
        click.echo(crayons.blue(str(e)), err=True)
        if 'no version found at all' in str(e):
            click.echo(
                crayons.blue(
                    'Please check your version specifier and version number. See PEP440 for more information.'
                )
            )
        raise RuntimeError

    return resolved_tree, resolver


def venv_resolve_deps(
    deps, which, project, pre=False, verbose=False, clear=False
):
    from .import resolver
    import json

    resolver = escape_grouped_arguments(resolver.__file__.rstrip('co'))
    cmd = '{0} {1} {2} {3} {4}'.format(
        escape_grouped_arguments(which('python')),
        resolver,
        '--pre' if pre else '',
        '--verbose' if verbose else '',
        '--clear' if clear else '',
    )
    os.environ['PIPENV_PACKAGES'] = '\n'.join(deps)
    c = delegator.run(cmd, block=True)
    del os.environ['PIPENV_PACKAGES']
    try:
        assert c.return_code == 0
    except AssertionError:
        if verbose:
            click.echo(c.out, err=True)
            click.echo(c.err, err=True)
        else:
            click.echo(c.err[int(len(c.err) / 2) - 1:], err=True)
        sys.exit(c.return_code)
    if verbose:
        click.echo(c.out.split('RESULTS:')[0], err=True)
    try:
        return json.loads(c.out.split('RESULTS:')[1].strip())

    except IndexError:
        raise RuntimeError('There was a problem with locking.')


def resolve_deps(
    deps,
    which,
    project,
    sources=None,
    verbose=False,
    python=False,
    clear=False,
    pre=False,
    allow_global=False,
):
    """Given a list of dependencies, return a resolved list of dependencies,
    using pip-tools -- and their hashes, using the warehouse API / pip9.
    """
    index_lookup = {}
    markers_lookup = {}
    python_path = which('python', allow_global=allow_global)
    backup_python_path = sys.executable
    results = []
    # First (proper) attempt:
    with HackedPythonVersion(python_version=python, python_path=python_path):
        try:
            resolved_tree, resolver = actually_resolve_reps(
                deps,
                index_lookup,
                markers_lookup,
                project,
                sources,
                verbose,
                clear,
                pre,
            )
        except RuntimeError:
            # Don't exit here, like usual.
            resolved_tree = None
    # Second (last-resort) attempt:
    if resolved_tree is None:
        with HackedPythonVersion(
            python_version='.'.join([str(s) for s in sys.version_info[:3]]),
            python_path=backup_python_path,
        ):
            try:
                # Attempt to resolve again, with different Python version information,
                # particularly for particularly particular packages.
                resolved_tree, resolver = actually_resolve_reps(
                    deps,
                    index_lookup,
                    markers_lookup,
                    project,
                    sources,
                    verbose,
                    clear,
                    pre,
                )
            except RuntimeError:
                sys.exit(1)
    for result in resolved_tree:
        if not result.editable:
            name = pep423_name(result.name)
            version = clean_pkg_version(result.specifier)
            index = index_lookup.get(result.name)
            if not markers_lookup.get(result.name):
                markers = str(
                    result.markers
                ) if result.markers and 'extra' not in str(
                    result.markers
                ) else None
            else:
                markers = markers_lookup.get(result.name)
            collected_hashes = []
            if any('python.org' in source['url'] or 'pypi.org' in source['url']
                   for source in sources):
                try:
                    # Grab the hashes from the new warehouse API.
                    r = requests.get(
                        'https://pypi.org/pypi/{0}/json'.format(name),
                        timeout=10,
                    )
                    api_releases = r.json()['releases']
                    cleaned_releases = {}
                    for api_version, api_info in api_releases.items():
                        cleaned_releases[
                            clean_pkg_version(api_version)
                        ] = api_info
                    for release in cleaned_releases[version]:
                        collected_hashes.append(release['digests']['sha256'])
                    collected_hashes = [
                        'sha256:' + s for s in collected_hashes
                    ]
                except (ValueError, KeyError, ConnectionError):
                    if verbose:
                        click.echo(
                            '{0}: Error generating hash for {1}'.format(
                                crayons.red('Warning', bold=True), name
                            )
                        )
            # Collect un-collectable hashes (should work with devpi).
            try:
                collected_hashes = collected_hashes + list(
                    list(resolver.resolve_hashes([result]).items())[0][1]
                )
            except (ValueError, KeyError, ConnectionError, IndexError):
                if verbose:
                    print('Error generating hash for {}'.format(name))
            collected_hashes = list(set(collected_hashes))
            d = {'name': name, 'version': version, 'hashes': collected_hashes}
            if index:
                d.update({'index': index})
            if markers:
                d.update({'markers': markers.replace('"', "'")})
            results.append(d)
    return results


def multi_split(s, split):
    """Splits on multiple given separators."""
    for r in split:
        s = s.replace(r, '|')
    return [i for i in s.split('|') if len(i) > 0]


def convert_deps_from_pip(dep):
    """"Converts a pip-formatted dependency to a Pipfile-formatted one."""
    dependency = {}
    req = get_requirement(dep)
    extras = {'extras': req.extras}
    # File installs.
    if (req.uri or req.path or is_installable_file(req.name)) and not req.vcs:
        # Assign a package name to the file, last 7 of it's sha256 hex digest.
        if not req.uri and not req.path:
            req.path = os.path.abspath(req.name)
        hashable_path = req.uri if req.uri else req.path
        req.name = hashlib.sha256(hashable_path.encode('utf-8')).hexdigest()
        req.name = req.name[len(req.name) - 7:]
        # {path: uri} TOML (spec 4 I guess...)
        if req.uri:
            dependency[req.name] = {'file': hashable_path}
        else:
            dependency[req.name] = {'path': hashable_path}
        if req.extras:
            dependency[req.name].update(extras)
        # Add --editable if applicable
        if req.editable:
            dependency[req.name].update({'editable': True})
    # VCS Installs. Extra check for unparsed git over SSH
    elif req.vcs or is_vcs(req.path):
        if req.name is None:
            raise ValueError(
                'pipenv requires an #egg fragment for version controlled '
                'dependencies. Please install remote dependency '
                'in the form {0}#egg=<package-name>.'.format(req.uri)
            )

        # Set up this requirement as a proper VCS requirement if it was not
        if not req.vcs and req.path.startswith(VCS_LIST):
            req.vcs = [vcs for vcs in VCS_LIST if req.path.startswith(vcs)][0]
            req.uri = '{0}'.format(req.path)
            req.path = None
        # Crop off the git+, etc part.
        if req.uri.startswith('{0}+'.format(req.vcs)):
            req.uri = req.uri[len(req.vcs) + 1:]
        dependency.setdefault(req.name, {}).update({req.vcs: req.uri})
        # Add --editable, if it's there.
        if req.editable:
            dependency[req.name].update({'editable': True})
        # Add subdirectory, if it's there
        if req.subdirectory:
            dependency[req.name].update({'subdirectory': req.subdirectory})
        # Add the specifier, if it was provided.
        if req.revision:
            dependency[req.name].update({'ref': req.revision})
        # Extras: e.g. #egg=requests[security]
        if req.extras:
            dependency[req.name].update({'extras': req.extras})
    elif req.extras or req.specs:
        specs = None
        # Comparison operators: e.g. Django>1.10
        if req.specs:
            r = multi_split(dep, '!=<>~')
            specs = dep[len(r[0]):]
            dependency[req.name] = specs
        # Extras: e.g. requests[socks]
        if req.extras:
            dependency[req.name] = extras
            if specs:
                dependency[req.name].update({'version': specs})
    # Bare dependencies: e.g. requests
    else:
        dependency[dep] = '*'
    # Cleanup when there's multiple values, e.g. -e.
    if len(dependency) > 1:
        for key in dependency.copy():
            if not hasattr(dependency[key], 'keys'):
                del dependency[key]
    return dependency


def is_star(val):
    return isinstance(val, six.string_types) and val == '*'


def is_pinned(val):
    return isinstance(val, six.string_types) and val.startswith('==')


def convert_deps_to_pip(deps, project=None, r=True, include_index=False):
    """"Converts a Pipfile-formatted dependency to a pip-formatted one."""
    dependencies = []
    for dep in deps.keys():
        # Default (e.g. '>1.10').
        extra = deps[dep] if isinstance(deps[dep], six.string_types) else ''
        version = ''
        index = ''
        # Get rid of '*'.
        if is_star(deps[dep]) or str(extra) == '{}':
            extra = ''
        hash = ''
        # Support for single hash (spec 1).
        if 'hash' in deps[dep]:
            hash = ' --hash={0}'.format(deps[dep]['hash'])
        # Support for multiple hashes (spec 2).
        if 'hashes' in deps[dep]:
            hash = '{0} '.format(
                ''.join(
                    [' --hash={0} '.format(h) for h in deps[dep]['hashes']]
                )
            )
        # Support for extras (e.g. requests[socks])
        if 'extras' in deps[dep]:
            extra = '[{0}]'.format(','.join(deps[dep]['extras']))
        if 'version' in deps[dep]:
            if not is_star(deps[dep]['version']):
                version = deps[dep]['version']
        # For lockfile format.
        if 'markers' in deps[dep]:
            specs = '; {0}'.format(deps[dep]['markers'])
        else:
            # For pipfile format.
            specs = []
            for specifier in specifiers:
                if specifier in deps[dep]:
                    if not is_star(deps[dep][specifier]):
                        specs.append(
                            '{0} {1}'.format(specifier, deps[dep][specifier])
                        )
            if specs:
                specs = '; {0}'.format(' and '.join(specs))
            else:
                specs = ''
        if include_index and not is_file(deps[dep]) and not is_vcs(deps[dep]):
            pip_src_args = []
            if 'index' in deps[dep]:
                pip_src_args = [project.get_source(deps[dep]['index'])]
            else:
                pip_src_args = project.sources
            pip_args = prepare_pip_source_args(pip_src_args)
            index = ' '.join(pip_args)
        # Support for version control
        maybe_vcs = [vcs for vcs in VCS_LIST if vcs in deps[dep]]
        vcs = maybe_vcs[0] if maybe_vcs else None
        # Support for files.
        if 'file' in deps[dep]:
            extra = '{1}{0}'.format(extra, deps[dep]['file']).strip()
            # Flag the file as editable if it is a local relative path
            if 'editable' in deps[dep]:
                dep = '-e '
            else:
                dep = ''
        # Support for paths.
        elif 'path' in deps[dep]:
            extra = '{1}{0}'.format(extra, deps[dep]['path']).strip()
            # Flag the file as editable if it is a local relative path
            if 'editable' in deps[dep]:
                dep = '-e '
            else:
                dep = ''
        if vcs:
            extra = '{0}+{1}'.format(vcs, deps[dep][vcs])
            # Support for @refs.
            if 'ref' in deps[dep]:
                extra += '@{0}'.format(deps[dep]['ref'])
            extra += '#egg={0}'.format(dep)
            # Support for subdirectory
            if 'subdirectory' in deps[dep]:
                extra += '&subdirectory={0}'.format(deps[dep]['subdirectory'])
            # Support for editable.
            if 'editable' in deps[dep]:
                # Support for --egg.
                dep = '-e '
            else:
                dep = ''
        s = '{0}{1}{2}{3}{4} {5}'.format(
            dep, extra, version, specs, hash, index
        ).strip(
        )
        dependencies.append(s)
    if not r:
        return dependencies

    # Write requirements.txt to tmp directory.
    f = tempfile.NamedTemporaryFile(suffix='-requirements.txt', delete=False)
    f.write('\n'.join(dependencies).encode('utf-8'))
    f.close()
    return f.name


def mkdir_p(newdir):
    """works the way a good mkdir should :)
        - already exists, silently complete
        - regular file in the way, raise an exception
        - parent directory(ies) does not exist, make them as well
        From: http://code.activestate.com/recipes/82465-a-friendly-mkdir/
    """
    if os.path.isdir(newdir):
        pass
    elif os.path.isfile(newdir):
        raise OSError(
            "a file with the same name as the desired dir, '{0}', already exists.".format(
                newdir
            )
        )

    else:
        head, tail = os.path.split(newdir)
        if head and not os.path.isdir(head):
            mkdir_p(head)
        if tail:
            os.mkdir(newdir)


def is_required_version(version, specified_version):
    """Check to see if there's a hard requirement for version
    number provided in the Pipfile.
    """
    # Certain packages may be defined with multiple values.
    if isinstance(specified_version, dict):
        specified_version = specified_version.get('version', '')
    if specified_version.startswith('=='):
        return version.strip() == specified_version.split('==')[1].strip()

    return True


def strip_ssh_from_git_uri(uri):
    """Return git+ssh:// formatted URI to git+git@ format"""
    if isinstance(uri, six.string_types):
        uri = uri.replace('git+ssh://', 'git+')
    return uri


def clean_git_uri(uri):
    """Cleans VCS uris from pip9 format"""
    if isinstance(uri, six.string_types):
        # Add scheme for parsing purposes, this is also what pip does
        if uri.startswith('git+') and '://' not in uri:
            uri = uri.replace('git+', 'git+ssh://')
    return uri


def is_vcs(pipfile_entry):
    import requirements

    """Determine if dictionary entry from Pipfile is for a vcs dependency."""
    if hasattr(pipfile_entry, 'keys'):
        return any(key for key in pipfile_entry.keys() if key in VCS_LIST)

    elif isinstance(pipfile_entry, six.string_types):
        return bool(
            requirements.requirement.VCS_REGEX.match(
                clean_git_uri(pipfile_entry)
            )
        )

    return False


def is_installable_file(path):
    """Determine if a path can potentially be installed"""
    from pip9.utils import is_installable_dir
    from pip9.utils.packaging import specifiers

    if hasattr(path, 'keys') and any(
        key for key in path.keys() if key in ['file', 'path']
    ):
        path = urlparse(path['file']).path if 'file' in path else path['path']
    if not isinstance(path, six.string_types) or path == '*':
        return False

    # If the string starts with a valid specifier operator, test if it is a valid
    # specifier set before making a path object (to avoid breaking windows)
    if any(path.startswith(spec) for spec in '!=<>~'):
        try:
            specifiers.SpecifierSet(path)
        # If this is not a valid specifier, just move on and try it as a path
        except specifiers.InvalidSpecifier:
            pass
        else:
            return False

    if not os.path.exists(os.path.abspath(path)):
        return False

    lookup_path = Path(path)
    absolute_path = '{0}'.format(lookup_path.absolute())
    if lookup_path.is_dir() and is_installable_dir(absolute_path):
        return True

    elif lookup_path.is_file() and is_archive_file(absolute_path):
        return True

    return False


def is_file(package):
    """Determine if a package name is for a File dependency."""
    if hasattr(package, 'keys'):
        return any(key for key in package.keys() if key in ['file', 'path'])

    if os.path.exists(str(package)):
        return True

    for start in SCHEME_LIST:
        if str(package).startswith(start):
            return True

    return False


def pep440_version(version):
    """Normalize version to PEP 440 standards"""
    from pip9.index import parse_version

    # Use pip built-in version parser.
    return str(parse_version(version))


def pep423_name(name):
    """Normalize package name to PEP 423 style standard."""
    name = name.lower()
    if any(i not in name for i in (VCS_LIST + SCHEME_LIST)):
        return name.replace('_', '-')

    else:
        return name


def proper_case(package_name):
    """Properly case project name from pypi.org."""
    # Hit the simple API.
    r = requests.get(
        'https://pypi.org/pypi/{0}/json'.format(package_name),
        timeout=0.3,
        stream=True,
    )
    if not r.ok:
        raise IOError(
            'Unable to find package {0} in PyPI repository.'.format(
                package_name
            )
        )

    r = parse.parse('https://pypi.org/pypi/{name}/json', r.url)
    good_name = r['name']
    return good_name


def split_section(input_file, section_suffix, test_function):
    """
    Split a pipfile or a lockfile section out by section name and test function

        :param dict input_file: A dictionary containing either a pipfile or lockfile
        :param str section_suffix: A string of the name of the section
        :param func test_function: A test function to test against the value in the key/value pair

    >>> split_section(my_lockfile, 'vcs', is_vcs)
    {
        'default': {
            "six": {
                "hashes": [
                    "sha256:832dc0e10feb1aa2c68dcc57dbb658f1c7e65b9b61af69048abc87a2db00a0eb",
                    "sha256:70e8a77beed4562e7f14fe23a786b54f6296e34344c23bc42f07b15018ff98e9"
                ],
                "version": "==1.11.0"
            }
        },
        'default-vcs': {
            "e1839a8": {
                "editable": true,
                "path": "."
            }
        }
    }
    """
    pipfile_sections = ('packages', 'dev-packages')
    lockfile_sections = ('default', 'develop')
    if any(section in input_file for section in pipfile_sections):
        sections = pipfile_sections
    elif any(section in input_file for section in lockfile_sections):
        sections = lockfile_sections
    else:
        # return the original file if we can't find any pipfile or lockfile sections
        return input_file

    for section in sections:
        split_dict = {}
        entries = input_file.get(section, {})
        for k in list(entries.keys()):
            if test_function(entries.get(k)):
                split_dict[k] = entries.pop(k)
        input_file['-'.join([section, section_suffix])] = split_dict
    return input_file


def split_file(file_dict):
    """Split VCS and editable dependencies out from file."""
    sections = {
        'vcs': is_vcs,
        'editable': lambda x: hasattr(x, 'keys') and x.get('editable'),
    }
    for k, func in sections.items():
        file_dict = split_section(file_dict, k, func)
    return file_dict


def merge_deps(
    file_dict,
    project,
    dev=False,
    requirements=False,
    ignore_hashes=False,
    blocking=False,
    only=False,
):
    """
    Given a file_dict, merges dependencies and converts them to pip dependency lists.
        :param dict file_dict: The result of calling :func:`pipenv.utils.split_file`
        :param :class:`pipenv.project.Project` project: Pipenv project
        :param bool dev=False: Flag indicating whether dev dependencies are to be installed
        :param bool requirements=False: Flag indicating whether to use a requirements file
        :param bool ignore_hashes=False:
        :param bool blocking=False:
        :param bool only=False:
        :return: Pip-converted 3-tuples of [deps, requirements_deps]
    """
    deps = []
    requirements_deps = []
    for section in list(file_dict.keys()):
        # Turn develop-vcs into ['develop', 'vcs']
        section_name, suffix = section.rsplit(
            '-', 1
        ) if '-' in section and not section == 'dev-packages' else (
            section, None
        )
        if not file_dict[section] or section_name not in (
            'dev-packages', 'packages', 'default', 'develop'
        ):
            continue

        is_dev = section_name in ('dev-packages', 'develop')
        if is_dev and not dev:
            continue

        if ignore_hashes:
            for k, v in file_dict[section]:
                if 'hash' in v:
                    del v['hash']
        # Block and ignore hashes for all suffixed sections (vcs/editable)
        no_hashes = True if suffix else ignore_hashes
        block = True if suffix else blocking
        include_index = True if not suffix else False
        converted = convert_deps_to_pip(
            file_dict[section], project, r=False, include_index=include_index
        )
        deps.extend((d, no_hashes, block) for d in converted)
        if dev and is_dev and requirements:
            requirements_deps.extend((d, no_hashes, block) for d in converted)
    return deps, requirements_deps


def recase_file(file_dict):
    """Recase file before writing to output."""
    if 'packages' in file_dict or 'dev-packages' in file_dict:
        sections = ('packages', 'dev-packages')
    elif 'default' in file_dict or 'develop' in file_dict:
        sections = ('default', 'develop')
    for section in sections:
        file_section = file_dict.get(section, {})
        # Try to properly case each key if we can.
        for key in list(file_section.keys()):
            try:
                cased_key = proper_case(key)
            except IOError:
                cased_key = key
            file_section[cased_key] = file_section.pop(key)
    return file_dict


def get_windows_path(*args):
    """Sanitize a path for windows environments

    Accepts an arbitrary list of arguments and makes a clean windows path"""
    return os.path.normpath(os.path.join(*args))


def find_windows_executable(bin_path, exe_name):
    """Given an executable name, search the given location for an executable"""
    requested_path = get_windows_path(bin_path, exe_name)
    if os.path.exists(requested_path):
        return requested_path

    # Ensure we aren't adding two layers of file extensions
    exe_name = os.path.splitext(exe_name)[0]
    files = [
        '{0}.{1}'.format(exe_name, ext) for ext in ['', 'py', 'exe', 'bat']
    ]
    exec_paths = [get_windows_path(bin_path, f) for f in files]
    exec_files = [
        filename for filename in exec_paths if os.path.isfile(filename)
    ]
    if exec_files:
        return exec_files[0]

    return find_executable(exe_name)


def get_converted_relative_path(path, relative_to= os.curdir):
    """Given a vague relative path, return the path relative to the given location"""
    return os.path.join('.', os.path.relpath(path, start=relative_to))


def walk_up(bottom):
    """Mimic os.walk, but walk 'up' instead of down the directory tree.
    From: https://gist.github.com/zdavkeos/1098474
    """
    bottom = os.path.realpath(bottom)
    # Get files in current dir.
    try:
        names = os.listdir(bottom)
    except Exception:
        return

    dirs, nondirs = [], []
    for name in names:
        if os.path.isdir(os.path.join(bottom, name)):
            dirs.append(name)
        else:
            nondirs.append(name)
    yield bottom, dirs, nondirs

    new_path = os.path.realpath(os.path.join(bottom, '..'))
    # See if we are at the top.
    if new_path == bottom:
        return

    for x in walk_up(new_path):
        yield x


def find_requirements(max_depth=3):
    """Returns the path of a Pipfile in parent directories."""
    i = 0
    for c, d, f in walk_up(os.getcwd()):
        i += 1
        if i < max_depth:
            if 'requirements.txt':
                r = os.path.join(c, 'requirements.txt')
                if os.path.isfile(r):
                    return r

    raise RuntimeError('No requirements.txt found!')




# Borrowed from pew to avoid importing pew which imports psutil
# See https://github.com/berdario/pew/blob/master/pew/_utils.py#L82
@contextmanager
def temp_environ():
    """Allow the ability to set os.environ temporarily"""
    environ = dict(os.environ)
    try:
        yield

    finally:
        os.environ.clear()
        os.environ.update(environ)


def is_valid_url(url):
    """Checks if a given string is an url"""
    pieces = urlparse(url)
    return all([pieces.scheme, pieces.netloc])


def download_file(url, filename):
    """Downloads file from url to a path with filename"""
    r = requests.get(url, stream=True)
    if not r.ok:
        raise IOError('Unable to download file')

    with open(filename, 'wb') as f:
        f.write(r.content)


def need_update_check():
    """Determines whether we need to check for updates."""
    mkdir_p(PIPENV_CACHE_DIR)
    p = os.sep.join((PIPENV_CACHE_DIR, '.pipenv_update_check'))
    if not os.path.exists(p):
        return True

    out_of_date_time = time() - (24 * 60 * 60)
    if os.path.isfile(p) and os.path.getmtime(p) <= out_of_date_time:
        return True

    else:
        return False


def touch_update_stamp():
    """Touches PIPENV_CACHE_DIR/.pipenv_update_check"""
    mkdir_p(PIPENV_CACHE_DIR)
    p = os.sep.join((PIPENV_CACHE_DIR, '.pipenv_update_check'))
    try:
        os.utime(p, None)
    except OSError:
        with open(p, 'w') as fh:
            fh.write('')


def normalize_drive(path):
    """Normalize drive in path so they stay consistent.

    This currently only affects local drives on Windows, which can be
    identified with either upper or lower cased drive names. The case is
    always converted to uppercase because it seems to be preferred.

    See: <https://github.com/pypa/pipenv/issues/1218>
    """
    if os.name != 'nt' or not isinstance(path, six.string_types):
        return path

    drive, tail = os.path.splitdrive(path)
    # Only match (lower cased) local drives (e.g. 'c:'), not UNC mounts.
    if drive.islower() and len(drive) == 2 and drive[1] == ':':
        return '{}{}'.format(drive.upper(), tail)

    return path


def is_readonly_path(fn):
    """Check if a provided path exists and is readonly.

    Permissions check is `bool(path.stat & stat.S_IREAD)` or `not os.access(path, os.W_OK)`
    """
    if os.path.exists(fn):
        return (os.stat(fn).st_mode & stat.S_IREAD) or not os.access(
            fn, os.W_OK
        )

    return False


def set_write_bit(fn):
    if os.path.exists(fn):
        os.chmod(fn, stat.S_IWRITE | stat.S_IWUSR)
    return


def rmtree(directory, ignore_errors=False):
    shutil.rmtree(
        directory, ignore_errors=ignore_errors, onerror=handle_remove_readonly
    )


def handle_remove_readonly(func, path, exc):
    """Error handler for shutil.rmtree.

    Windows source repo folders are read-only by default, so this error handler
    attempts to set them as writeable and then proceed with deletion."""
    # Check for read-only attribute
    default_warning_message = 'Unable to remove file due to permissions restriction: {!r}'
    # split the initial exception out into its type, exception, and traceback
    exc_type, exc_exception, exc_tb = exc
    if is_readonly_path(path):
        # Apply write permission and call original function
        set_write_bit(path)
        try:
            func(path)
        except (OSError, IOError) as e:
            if e.errno in [errno.EACCES, errno.EPERM]:
                warnings.warn(
                    default_warning_message.format(path), ResourceWarning
                )
                return

    if exc_exception.errno in [errno.EACCES, errno.EPERM]:
        warnings.warn(default_warning_message.format(path), ResourceWarning)
        return

    raise


class TemporaryDirectory(object):
    """Create and return a temporary directory.  This has the same
    behavior as mkdtemp but can be used as a context manager.  For
    example:

        with TemporaryDirectory() as tmpdir:
            ...

    Upon exiting the context, the directory and everything contained
    in it are removed.
    """

    def __init__(self, suffix=None, prefix=None, dir=None):
        if 'RAM_DISK' in os.environ:
            import uuid

            name = uuid.uuid4().hex
            dir_name = os.path.sep.join([os.environ['RAM_DISK'].strip(), name])
            os.mkdir(dir_name)
            self.name = dir_name
        else:
            self.name = tempfile.mkdtemp(suffix, prefix, dir)
        self._finalizer = finalize(
            self,
            self._cleanup,
            self.name,
            warn_message="Implicitly cleaning up {!r}".format(self),
        )

    @classmethod
    def _cleanup(cls, name, warn_message):
        rmtree(name)
        warnings.warn(warn_message, ResourceWarning)

    def __repr__(self):
        return "<{} {!r}>".format(self.__class__.__name__, self.name)

    def __enter__(self):
        return self

    def __exit__(self, exc, value, tb):
        self.cleanup()

    def cleanup(self):
        if self._finalizer.detach():
            rmtree(self.name)



# -*- coding: utf-8 -*-

from prettytoml.lexer import _munch_a_token
from prettytoml.lexer import *

# A mapping from token types to a sequence of pairs of (source_text, expected_matched_text)
valid_tokens = {
    tokens.TYPE_COMMENT: (
        (
            '# My very insightful comment about the state of the universe\n# And now for something completely different!',
            '# My very insightful comment about the state of the universe',
        ),
    ),
    tokens.TYPE_STRING: (
        ('"a valid hug3 text" "some other string" = 42', '"a valid hug3 text"'),
        (
            r'"I\'m a string. \"You can quote me\". Name\tJos\u00E9\nLocation\tSF." "some other string" = 42',
            r'"I\'m a string. \"You can quote me\". Name\tJos\u00E9\nLocation\tSF."'
        ),
        ('"ʎǝʞ" key', '"ʎǝʞ"'),
        ('""', '""'),
        ('"t"', '"t"'),
    ),
    tokens.TYPE_MULTILINE_STRING: (
        ('"""\nRoses are red\nViolets are blue""" """other text"""', '"""\nRoses are red\nViolets are blue"""'),
    ),
    tokens.TYPE_LITERAL_STRING: (
        (r"'This is \ \n a \\ literal string' 'another \ literal string'", r"'This is \ \n a \\ literal string'"),
    ),
    tokens.TYPE_MULTILINE_LITERAL_STRING: (
        (
            "'''\nThe first newline is\ntrimmed in raw strings.\n   All other whitespace\n   is preserved.\n''' '''some other\n\n\t string'''",
            "'''\nThe first newline is\ntrimmed in raw strings.\n   All other whitespace\n   is preserved.\n'''"
        ),
    ),
    tokens.TYPE_DATE: (
        ('1979-05-27 5345', '1979-05-27'),
        ('1979-05-27T07:32:00Z something', '1979-05-27T07:32:00Z'),
        ('1979-05-27T00:32:00-07:00 ommm', '1979-05-27T00:32:00-07:00'),
        ('1979-05-27T00:32:00.999999-07:00 2346', '1979-05-27T00:32:00.999999-07:00'),
    ),
    tokens.TYPE_WHITESPACE: (
        (' \t\n \r  some_text', ' '),
    ),
    tokens.TYPE_INTEGER: (
        ('+99 "number"', "+99"),
        ('42 fwfwef', "42"),
        ('-17 fh34g34g', "-17"),
        ('5_349_221 apples', "5_349_221"),
        ('-1_2_3_4_5 steps', '-1_2_3_4_5')
    ),
    tokens.TYPE_FLOAT: (
        ('1.0 fwef', '1.0'),
        ('3.1415 g4g', '3.1415'),
        ('-0.01 433re', '-0.01'),
        ('5e+2_2 ersdvf', '5e+2_2'),
        ('1e6 ewe23', '1e6'),
        ('-2E-2.2 3 rf23', '-2E-2'),
        ('6.626e-34 +234f', '6.626e-34'),
        ('9_224_617.445_991_228_313 f1ewer 23f4h = nonesense', '9_224_617.445_991_228_313'),
        ('1e1_000 2346f,ef2!!', '1e1_000'),
    ),
    tokens.TYPE_BOOLEAN: (
        ('false business = true', 'false'),
        ('true true', 'true'),
    ),
    tokens.TYPE_OP_SQUARE_LEFT_BRACKET: (
        ('[table_name]', '['),
    ),
    tokens.TYPE_OP_SQUARE_RIGHT_BRACKET: (
        (']\nbusiness = awesome', ']'),
    ),
    tokens.TYPE_OP_CURLY_LEFT_BRACKET: (
        ('{item_exists = no}', '{'),
    ),
    tokens.TYPE_OP_CURLY_RIGHT_BRACKET: (
        ('} moving on', '}'),
    ),
    tokens.TYPE_OP_COMMA: (
        (',item2,item4', ','),
    ),
    tokens.TYPE_OP_ASSIGNMENT: (
        ('== 42', '='),
    ),
    tokens.TYPE_OP_DOUBLE_SQUARE_LEFT_BRACKET: (
        ('[[array.of.tables]]', '[['),
    ),
    tokens.TYPE_OP_DOUBLE_SQUARE_RIGHT_BRACKET: (
        (']] item=3', ']]'),
    ),
    tokens.TYPE_BARE_STRING: (
        ('key another', 'key'),
        ('bare_key 2fews', 'bare_key'),
        ('bare-key kfcw', 'bare-key'),
    ),
    tokens.TYPE_OPT_DOT: (
        ('."another key"', '.'),
        ('.subname', '.'),
    ),
    tokens.TYPE_NEWLINE: (
        ('\n\r \n', '\n'),
    )
}

# A mapping from a token type to a sequence of (source, matched_text) pairs that shouldn't result from consuming the
# source text.
invalid_tokens = {
    tokens.TYPE_INTEGER: (
        ('_234_423', ''),
        ('0446234234', ''),
    ),
    tokens.TYPE_STRING: (
        ('"""', '"""'),
    ),
    tokens.TYPE_BOOLEAN: (
        ('True', 'True'),
        ('True', 'true'),
    ),
    tokens.TYPE_FLOAT: (
        ('', ''),
    )
}


def test_valid_tokenizing():
    for token_type in valid_tokens:
        for (source, expected_match) in valid_tokens[token_type]:

            token = _munch_a_token(source)
            assert token, "Failed to tokenize: {}\nExpected: {}\nOut of: {}\nGot nothing!".format(
                token_type, expected_match, source)

            assert token.type == token_type, \
                "Expected type: {}\nOut of: {}\nThat matched: {}\nOf type: {}".format(
                    token_type, source, token.source_substring, token.type)
            assert token.source_substring == expected_match


def test_invalid_tokenizing():
    for token_type in invalid_tokens:
        for source, expected_match in invalid_tokens[token_type]:
            token = _munch_a_token(source)
            if token:
                assert not (token.type == token_type and token.source_substring == expected_match)


def test_token_type_order():
    type_a = tokens.TokenType('a', 5, is_metadata=False)
    type_b = tokens.TokenType('b', 0, is_metadata=False)
    type_c = tokens.TokenType('c', 3, is_metadata=False)

    assert type_b < type_c < type_a
    assert type_a > type_c > type_b



from __future__ import absolute_import

import os.path
import tempfile

from pip9.utils import rmtree


class BuildDirectory(object):

    def __init__(self, name=None, delete=None):
        # If we were not given an explicit directory, and we were not given an
        # explicit delete option, then we'll default to deleting.
        if name is None and delete is None:
            delete = True

        if name is None:
            # We realpath here because some systems have their default tmpdir
            # symlinked to another directory.  This tends to confuse build
            # scripts, so we canonicalize the path by traversing potential
            # symlinks here.
            name = os.path.realpath(tempfile.mkdtemp(prefix="pip-build-"))
            # If we were not given an explicit directory, and we were not given
            # an explicit delete option, then we'll default to deleting.
            if delete is None:
                delete = True

        self.name = name
        self.delete = delete

    def __repr__(self):
        return "<{} {!r}>".format(self.__class__.__name__, self.name)

    def __enter__(self):
        return self.name

    def __exit__(self, exc, value, tb):
        self.cleanup()

    def cleanup(self):
        if self.delete:
            rmtree(self.name)
